use core::fmt;

use base64::{prelude::BASE64_STANDARD, Engine};
use sha2::Digest;

#[derive(Debug, PartialEq)]
pub enum Error {
    InvalidSRIFormat,
    AlgorithmNotSupported,
    UnableToDecode(base64::DecodeError),
}

#[derive(Debug, PartialEq)]
pub enum Algorithm {
    SHA256,
    SHA384,
    SHA512,
}

impl Algorithm {
    pub fn digest(&self, data: impl AsRef<[u8]>) -> Vec<u8> {
        match self {
            Self::SHA256 => sha2::Sha256::digest(data).to_vec(),
            Self::SHA384 => sha2::Sha384::digest(data).to_vec(),
            Self::SHA512 => sha2::Sha512::digest(data).to_vec(),
        }
    }
}

impl fmt::Display for Algorithm {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        let value = match self {
            Self::SHA256 => "sha256",
            Self::SHA384 => "sha384",
            Self::SHA512 => "sha512",
        };

        write!(f, "{value}")
    }
}

impl TryFrom<&str> for Algorithm {
    type Error = Error;

    fn try_from(value: &str) -> Result<Self, Self::Error> {
        match value {
            "sha256" => Ok(Self::SHA256),
            "sha384" => Ok(Self::SHA384),
            "sha512" => Ok(Self::SHA512),
            _ => Err(Error::AlgorithmNotSupported),
        }
    }
}

#[derive(Debug)]
pub struct Integrity {
    algorithm: Algorithm,
    hash: Vec<u8>,
}

impl Integrity {
    pub fn new(algorithm: Algorithm, hash: impl Into<Vec<u8>>) -> Self {
        Self {
            algorithm,
            hash: hash.into(),
        }
    }

    pub fn from_base64(algorithm: Algorithm, base64: &str) -> Result<Self, Error> {
        Ok(Self {
            algorithm,
            hash: Self::b64_to_hex(base64).map_err(Error::UnableToDecode)?,
        })
    }

    fn b64_to_hex(s: &str) -> Result<Vec<u8>, base64::DecodeError> {
        BASE64_STANDARD.decode(s)
    }

    pub fn verify(&self, data: impl AsRef<[u8]>) -> bool {
        self.algorithm.digest(data) == self.hash
    }

    fn split(value: &str) -> Result<(&str, &str), Error> {
        value.split_once('-').ok_or(Error::InvalidSRIFormat)
    }
}

impl fmt::Display for Integrity {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(
            f,
            "{}-{}",
            self.algorithm,
            BASE64_STANDARD.encode(self.hash.as_slice())
        )
    }
}

impl TryFrom<&str> for Integrity {
    type Error = Error;

    fn try_from(value: &str) -> Result<Self, Self::Error> {
        let (algo, hash) = Self::split(value)?;
        println!("{algo} {hash}");
        Self::from_base64(Algorithm::try_from(algo)?, hash)
    }
}

impl TryFrom<String> for Integrity {
    type Error = Error;

    fn try_from(value: String) -> Result<Self, Self::Error> {
        let (algo, hash) = Self::split(&value)?;
        Self::from_base64(Algorithm::try_from(algo)?, hash)
    }
}

#[cfg(test)]
mod test {
    use crate::Algorithm;
    use crate::Error;
    use crate::Integrity;
    use base64::Engine;
    use sha2::Digest;

    #[test]
    fn try_from_sha256() {
        let value = "alert('Hello, world.');";
        let digest = sha2::Sha256::digest(value);
        let hash = base64::prelude::BASE64_STANDARD.encode(digest);
        let integrity_value = format!("sha256-{hash}");
        let integrity = Integrity::try_from(integrity_value.as_str()).unwrap();

        assert_eq!(integrity.algorithm, Algorithm::SHA256);
        assert_eq!(integrity.hash, digest.as_slice());
        assert_eq!(integrity.to_string(), integrity_value)
    }

    #[test]
    fn try_from_sha384() {
        let value = "alert('Hello, world.');";
        let digest = sha2::Sha384::digest(value);
        let hash = base64::prelude::BASE64_STANDARD.encode(digest);
        let integrity_value = format!("sha384-{hash}");
        let integrity = Integrity::try_from(integrity_value.as_str()).unwrap();

        assert_eq!(integrity.algorithm, Algorithm::SHA384);
        assert_eq!(integrity.hash, digest.as_slice());
        assert_eq!(integrity.to_string(), integrity_value);
        assert!(integrity.verify(value));
    }

    #[test]
    fn try_from_sha512() {
        let value = "alert('Hello, world.');";
        let digest = sha2::Sha512::digest(value);
        let hash = base64::prelude::BASE64_STANDARD.encode(digest);
        let integrity_value = format!("sha512-{hash}");
        let integrity = Integrity::try_from(integrity_value.as_str()).unwrap();

        assert_eq!(integrity.algorithm, Algorithm::SHA512);
        assert_eq!(integrity.hash, digest.as_slice());
        assert_eq!(integrity.to_string(), integrity_value);
        assert!(integrity.verify(value));
    }

    #[test]
    fn bad_format() {
        let values = [
            "sha384",
            "sha384=o9w34t98ywhutkjwgjiaroakl",
            "ml,goxu904t8y9 sha256olkwtgw,.rt gmwl tow3t",
            "sha512,wa.rtgtliworh gtnkrlgtmw",
        ];

        let invalid_sri = values.map(Integrity::try_from);

        for integrity in invalid_sri {
            assert_eq!(integrity.unwrap_err(), Error::InvalidSRIFormat);
        }

        let undecodable = "sha384-AAAAAAA";
        let integrity = Integrity::try_from(undecodable);
        assert!(matches!(integrity.unwrap_err(), Error::UnableToDecode(_)));
    }

    #[test]
    fn unsupported_hash() {
        let value = "alert('Hello, world.');";
        let digest = sha2::Sha224::digest(value);
        let hash = base64::prelude::BASE64_STANDARD.encode(digest);
        let integrity_value = format!("sha224-{hash}");
        let integrity = Integrity::try_from(integrity_value.as_str());
        assert_eq!(integrity.unwrap_err(), Error::AlgorithmNotSupported);
    }
}
